-- Allows you to keep track of people you have had whisper conversations with

-- Variables
TellTrackMessageDB = { }
local TellTrack_WhisperMessages = { }
local TellTrack_OfflinePlayers = { }

-- hooked funtions
local SavedBNSendWhisper = nil
local SavedSendChatMessage = nil
local SavedFCFDock_SelectWindow = nil
local SavedFCF_SelectDockFrame = nil
local SavedChatFrame_OpenChat = nil
local SavedFCF_Close = nil
local SavedFCF_OpenNewWindow = nil

SELECTED_CHAT_FRAME = DEFAULT_CHAT_FRAME

function TellTrack:OnLoadCustom( frame )
	frame:RegisterEvent("PLAYER_ENTERING_WORLD")

	self:AddSlashCommand( nil, "/ttrack" )
	SlashCmdList["TELLTRACKRETELL"] = TellTrack_WhisperToPreviousTarget
	self:AddSlashCommand( "RETELL", "/retell", true )
	self:AddSlashCommand( "RETELL", "/re" )
end

function TellTrack:OnPlayerLogin( frame )
	frame:RegisterEvent("CHAT_MSG_WHISPER")
	frame:RegisterEvent("CHAT_MSG_WHISPER_INFORM")
	frame:RegisterEvent("CHAT_MSG_BN_WHISPER")
	frame:RegisterEvent("CHAT_MSG_BN_WHISPER_INFORM")
	frame:RegisterEvent("CHAT_MSG_SYSTEM")
	frame:RegisterEvent("CHAT_MSG_BN_WHISPER_PLAYER_OFFLINE")
	frame:RegisterEvent("CHAT_MSG_BN_INLINE_TOAST_ALERT")

	local db,o = self._db,self._options
	local _inst = self._acelib
	self.L = _inst.L
	self.LR = _inst.LR

	-- init vars
	o.Realm,o.Race,o.Player,o.Faction = (GetRealmName()),(UnitRace("player")),(UnitName("Player")),(UnitFactionGroup("Player"))
	o.RealmPostfix = "-" .. o.Realm
	o.RealmPostfixCmp = self:LowerUTF8(o.RealmPostfix)
	o.ListProfile = o.Realm .. " - " .. o.Faction
	o.PlayerProfile = o.Player .. " - " .. o.Realm

	-- curent conversation
	self._current = { name="", showAll=true, }

	-- load player list
	if o.ListSave ~= 1 then db.Log = { } end
	self._list = db.Log[o.ListProfile] or { }
	db.Log[o.ListProfile] = self._list

	-- load frame list
	self._chat = db.Chat[o.PlayerProfile] or { }
	db.Chat[o.PlayerProfile] = self._chat

	-- load messages
	self._messages = TellTrack_WhisperMessages
	self:EnsureMessageDB()
	for i,v in pairs(TellTrackMessageDB) do tinsert(self._messages, v) end

	-- update cmds and menus
	if _inst and _inst.menu then			-- remove default help option
		local args = _inst.menu.args
		local main = args.main; main.type = "toggle"; main.set = main.func;
			main.get = function() return TellTrackFrame:IsVisible() end
		args.options.passValue = "help"
		args.help.passValue = "info"
	end

	-- hook chat functions
	if BNSendWhisper ~= SavedBNSendWhisper then						-- Hook the chat event handler 'before'
		SavedBNSendWhisper = BNSendWhisper
		BNSendWhisper = TellTrack_BNSendWhisper
	end
	if SendChatMessage ~= SavedSendChatMessage then					-- Hook the chat event handler 'before'
		SavedSendChatMessage = SendChatMessage
		SendChatMessage = TellTrack_SendChatMessage
	end
	if FCFDock_SelectWindow ~= TellTrack_FCFDock_SelectWindow then	-- Hook dock frame selection 'after'
		SavedFCFDock_SelectWindow = FCFDock_SelectWindow
		FCFDock_SelectWindow = TellTrack_FCFDock_SelectWindow
	end
	if FCF_SelectDockFrame ~= TellTrack_FCF_SelectDockFrame then	-- Hook dock frame selection 'after'
		SavedFCF_SelectDockFrame = FCF_SelectDockFrame
		FCF_SelectDockFrame = TellTrack_FCF_SelectDockFrame
	end
	if ChatFrame_OpenChat ~= TellTrack_ChatFrame_OpenChat then		-- Hook open chat to open as whisper
		SavedChatFrame_OpenChat = ChatFrame_OpenChat
		ChatFrame_OpenChat = TellTrack_ChatFrame_OpenChat
	end
	if FCF_Close ~= TellTrack_FCF_Close then						-- Hook open chat to open as whisper
		SavedFCF_Close = FCF_Close
		FCF_Close = TellTrack_FCF_Close
	end
	if FCF_OpenNewWindow ~= TellTrack_FCF_OpenNewWindow then		-- Hook open chat to open as whisper
		SavedFCF_OpenNewWindow = FCF_OpenNewWindow
		FCF_OpenNewWindow = TellTrack_FCF_OpenNewWindow
	end
end

-- due to bnet lag and wow not re-render ui for the |K
function TellTrack:OnPlayerEnteringWorld( )
	local o,_chat = self._options,self._chat

	TT_Button:OnAction(nil, "enable", o.Enable)
	TT_Button:OnSizeChanged( TellTrackFrame )
	TT_Button:UpdateBorderVisibility()

	_chat.title = select(2,self:DispName(_chat.name))
	self._bnsync = _chat.title == UNKNOWN and (self._bnsync or 1) or nil
	self:Debug("_bnsync",self._bnsync and ""..self._bnsync or "is synced", "name",_chat.title)
	if self._bnsync then return end

	local frame = _chat.frame and _G[_chat.frame]
	if frame then FCF_SetWindowName(frame, _chat.title) end
	self:RedrawWhisperFrame()
	self:UpdateTellTrackButtonsText()
end

function TellTrack:UpdateFrame_OnUpdate()
	if not self._bnsync then return end 
	if self._bnsync > 100 then self.UpdateFrame_OnUpdate=nil end	-- try for ~25 seconds
	self._bnsync = self._bnsync + 1
	self:OnPlayerEnteringWorld()
end

function TellTrack:OnCommandCustom( msg, cmd, args )	--self:Debug("OnCommandCustom", msg, cmd, args)
	local o,L = self._options, self.L

	if not cmd then
		TT_Button:OnAction(nil, "enable")
		return true
	elseif TT_Button:OnAction(nil, cmd, args) then
		return true
	elseif cmd == "help" then
		self:Print(TT_CHAT_USEAGE)
		local list,short = TellTrack_cmdlist, self._meta.short
		for k,v in pairs(list) do
			self:Print("/"..short.." "..k.." - ".. L[v.help])
		end
	end
end

function TellTrack:Print_Help()
	self:Print(TT_CHAT_INFO)
	self:Print(TT_CHAT_INFO1)
	self:Print(TT_CHAT_INFO2)
	self:Print(TT_CHAT_INFO3)
end


function TellTrack_WhisperToPreviousTarget(msg)
	local o = TellTrack._options
	if not o.LastTell then return end
	SendChatMessage(msg, TellTrack:NameType(o.LastTell), o.LastLang, o.LastTell)
end


-- Handles events
function TellTrack:OnEventCustom(frame, event, ...)
	local ar1,ar2,ar3,ar4,ar5,ar6,_,_,_,_,ar11,_,ar13 = select(1, ...)
	--self:Debug("TellTrack:OnEventCustom", event, ar1, ar2, ar3, ar4, ar5, ar6, ar11, ar13)

	if event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM" or
		event == "CHAT_MSG_BN_WHISPER" or event == "CHAT_MSG_BN_WHISPER_INFORM" then

		self:HookAllAddMessages()
		if type(ar1) ~= "string" then return end

		--Special Hidden Whisper Ignores
		local len = ar1:len()
		if Guilded and len > 2 and ar1:sub(1,2) == "<G" and (len ~= 4 or ar1:sub(1,4) ~= "<GM>") then return	--Hide Guilded messages
		elseif GuildAds and len > 3 and ar1:sub(1,3) == "<GA" then return				--Hide GuildAds messages
		elseif len > 3 and ar1:sub(1,3) == "<T>" then return							--Hide Telepathy messages
		elseif len > 1 and ar1:sub(1,1) == "/" then return								--Hide AceComm messages
		elseif len > 7 and ar1:sub(1,7) == "SKP1019" then return						--Hide SKP messages
		end

		local etype = string.sub(event, 10)
		local message = gsub(ar1, "%%", "%%%%")

		if etype == "WHISPER" or etype == "BN_WHISPER" then
			self:HandleMessageSentOrRecieved(ar2, false)
		end

		local name,cmpName = (self:NormName(ar2)),(self:CmpName(ar2))
		if cmpName then
			TellTrack_OfflinePlayers[cmpName] = nil
		end

		--self:Debug("TellTrack:OnEventCustom", "type",etype, "tag",ar6, "ar2",ar2, "name",name, "msg",message, "time",self:GetTime(nil, self._update.timediff))
		local whisperItem = { type=etype, tag=ar6, name=name, msg=message, time=self:GetTime(nil, self._update.timediff), }

		tinsert(TellTrack_WhisperMessages, whisperItem)
		tinsert(TellTrackMessageDB, whisperItem)
		self:EnsureMessageDB()

	elseif event == "CHAT_MSG_SYSTEM" or event == "CHAT_MSG_BN_WHISPER_PLAYER_OFFLINE"
		or event == "CHAT_MSG_BN_INLINE_TOAST_ALERT" then

		if type(ar1) ~= "string" then return end

		local result,name = false,nil;
		if result==false then if ar1=="FRIEND_OFFLINE" then name=ar2; result=true end end
		if result==false then if ar1=="FRIEND_ONLINE" then name=ar2; result=nil end end
		if result==false then _,_,name = strfind(ar1, TT_ERR_CHAT_PLAYER_NOT_FOUND_S); if name then result=true end end
		if result==false then _,_,name = strfind(ar1, TT_ERR_FRIEND_OFFLINE_S); if name then result=true end end
		if result==false then _,_,name = strfind(ar1, TT_ERR_FRIEND_ONLINE_SS); if name then result=nil end end

		if name then
			TellTrack_OfflinePlayers[self:CmpName(name)] = result
			self:UpdateTellTrackButtonsText()
		end

	elseif event == "PLAYER_ENTERING_WORLD" then
		frame:UnregisterEvent("PLAYER_ENTERING_WORLD")
		if self.OnPlayerEnteringWorld then self:OnPlayerEnteringWorld(frame) end

	end
end

-- adds a whisper to a frame
function TellTrack:EnsureMessageDB(count)
	local o = self._options
	count = count or (o.MessageSave == 1 and o.MessageCount or 0)
	while #TellTrackMessageDB > count do table.remove(TellTrackMessageDB, 1) end
end

function TellTrack:RedrawWhisperFrame( )		--self:Debug("RedrawWhisperFrame", self._current and  self._current.name)
	local _chat,_current = self._chat,self._current
	local frame = _chat.frame and _G[_chat.frame]
	if not frame then return end

	local scrollOffset,found = frame:GetScrollOffset(),false
	--local scrollOffset,found = frame:GetCurrentScroll(),false
	frame:Clear()

	--Display only the valid whispers.
	local cmpCurName = self:CmpName(_current.name)
	for i,v in pairs(TellTrack_WhisperMessages) do
		if _current.showAll or self:CmpName(v.name) == cmpCurName then
			local info = ChatTypeInfo[v.type]
			local name,normNameL = self:DispName(v.name)
			local part1 = _G["CHAT_"..v.type.."_GET"]..v.msg
			local part2 = (v.tag and _G["CHAT_FLAG_"..v.tag] or "").."|Hplayer:"..normNameL.."|h".."["..normNameL.."]".."|h"
			if TasteTheNaimbow_Loaded then
				part2 = "|cff"..TasteTheNaimbowExternalColor(normNameL)..part2.."|r"
			end
			frame:AddMessage(format(part1, part2), info.r, info.g, info.b, info.id, nil, v.time)
			found = true
		end
	end

	frame:SetScrollOffset(scrollOffset)
	return found
end

function TellTrack:ShowConversationWith(name, showAll)		--self:Debug("ShowConversationWith", name, showAll)
	local db,o,_chat = self._db,self._options,self._chat
	local frame = self:EnsureChatFrame()
	if not frame then return end

	local normName,dispName,dispNameL = (self:NormName(name)),self:DispName(name)
	self._current = { name=normName, showAll=showAll, }

	if not self:RedrawWhisperFrame() then
		self._current.showAll = true
		local timeStamp = o.TimeStamps
		o.TimeStamps = 0
		frame:AddMessage(format(self.L["-- No Current Conversation with %s --"], dispNameL))
		o.TimeStamps = timeStamp
	end

	-- Show the frame and tab
	FCF_SetWindowName(frame, dispNameL)
	_chat.name = name
	_chat.title = dispNameL
	self:SelectChatFrame(frame)
	--if frame.isDocked then
	--	FCF_SelectDockFrame(frame)
	--end
end

function TellTrack:ShowMainChatFrame()
	local frame = DEFAULT_CHAT_FRAME
	if frame.isDocked then
		FCF_SelectDockFrame(frame)
	end
end

function TellTrack_ChatFrame_AddMessage(frame, text, r, g, b, id, addToStart, ts)
	local tt = TellTrack
	local AddMessage = SavedChatFrame_AddMessages[frame:GetName()]

	if frame:GetName() ~= tt._chat.frame then
		AddMessage(frame, text, r, g, b, id, addToStart)
		return					-- incorrect whisper frame
	end

	local stamps = ChatTimeStamps_ENABLED
	ChatTimeStamps_ENABLED = 0
	if not text or strlen(text) < 5 then
		AddMessage(frame, text, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end

	local timedText = text
	if tt._options.TimeStamps == 1 then
		timedText = tt:GetTimeStamp(ts)..text
	end
	if not tt._current or tt._current.showAll then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end

	local name = tt._current.name
	local normName,normNameL = tt:NormName(name)
	local dispName,dispNameL = tt:DispName(name)

	local namecolor = "ffffff"
	if TasteTheNaimbow_Loaded and TasteTheNaimbowExternalColor then
		namecolor = self:SafeCall(TasteTheNaimbowExternalColor, normNameL)
	end

	local Hplayer = "|Hplayer:"..dispName.."|h["..dispName.."]|h"
	local HplayerL = "[|Hplayer:"..dispName.."|h"..dispName.."|h]"
	local HplayerLocal = "|Hplayer:"..dispNameL.."|h["..dispNameL.."]|h"
	local HplayerLocalL = "[|Hplayer:"..dispNameL.."|h"..dispNameL.."|h]"
	local HplayerMixed = "|Hplayer:"..dispName.."|h["..dispNameL.."]|h"
	local HplayerMixedL = "[|Hplayer:"..dispName.."|h"..dispNameL.."|h]"
	--DEFAULT_CHAT_FRAME:AddMessage("Real "..string.gsub(string.gsub(text,":%d+\124","\124"),"\124","\\124"))
	--DEFAULT_CHAT_FRAME:AddMessage("Fake "..string.gsub(format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"),"\124","\\124"))
	--DEFAULT_CHAT_FRAME:AddMessage(tostring((tt:startsWith(string.gsub(text,":%d+\124","\124"), format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h")))))
	--local matchName = strlower(normName:gsub("\-","\-"))
	local sm = strlower(text)
		sm = sm:gsub(":%d+|","|"):gsub(":%d+:",":"):gsub(":%d+|","|"):gsub(":%d+:",":")	-- remove numbers
		sm = sm:gsub("|hbnplayer:","|hplayer:"):gsub(":bn_whisper:",":whisper:")		-- normalize bn
		sm = sm:gsub(":whisper:[^|^:]+",""):gsub(":whisper:.+|h%[","|h[")				-- remove verbose command
	local smlen = sm:len()
	--print("1", text and strlower(text:gsub("|","!")))
	--print("2", strlower(sm:gsub("|","!")))
	--print("3", strlower(format(CHAT_WHISPER_GET, Hplayer):gsub("|","!")))

	local function match(get, name)
		local s = strlower(format(get, name))
		return sm == s or (smlen > s:len() and sm:sub(1, s:len()) == s)
	end
	if	-- this whole section, above and below, needs much love, not most efficent.  need early outs, etc...
		match(CHAT_WHISPER_GET, Hplayer) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..Hplayer) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..Hplayer) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..Hplayer) or
		match(CHAT_WHISPER_INFORM_GET, Hplayer) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..Hplayer.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..Hplayer.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..Hplayer.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..Hplayer.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..Hplayer.."|r") or
		match(CHAT_WHISPER_GET, HplayerL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerL) or
		match(CHAT_WHISPER_INFORM_GET, HplayerL) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerL.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerL.."|r") or

		match(CHAT_WHISPER_GET, HplayerLocal) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerLocal) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerLocal) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerLocal) or
		match(CHAT_WHISPER_INFORM_GET, HplayerLocal) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerLocal.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerLocal.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerLocal.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerLocal.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerLocal.."|r") or
		match(CHAT_WHISPER_GET, HplayerLocalL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerLocalL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerLocalL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerLocalL) or
		match(CHAT_WHISPER_INFORM_GET, HplayerLocalL) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerLocalL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerLocalL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerLocalL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerLocalL.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerLocalL.."|r") or

		match(CHAT_WHISPER_GET, HplayerMixed) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerMixed) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerMixed) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerMixed) or
		match(CHAT_WHISPER_INFORM_GET, HplayerMixed) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerMixed.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerMixed.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerMixed.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerMixed.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerMixed.."|r") or
		match(CHAT_WHISPER_GET, HplayerMixedL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerMixedL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerMixedL) or
		match(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerMixedL) or
		match(CHAT_WHISPER_INFORM_GET, HplayerMixedL) or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerMixedL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerMixedL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerMixedL.."|r") or
		match(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerMixedL.."|r") or
		match(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerMixedL.."|r")
	then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
	end
end


function TellTrack:HookAllAddMessages()
	if SavedChatFrame_AddMessages then return end

	SavedChatFrame_AddMessages = {}
	for i=1, NUM_CHAT_WINDOWS do
		local frameName = "ChatFrame"..i
		local frame = _G[frameName]
		if frame and (frame.AddMessage ~= TellTrack_ChatFrame_AddMessage) then
			SavedChatFrame_AddMessages[frameName] = frame.AddMessage
			frame.AddMessage = TellTrack_ChatFrame_AddMessage
		end
	end
end

function TellTrack:FindChatFrame( ... )		--self:Debug("FindChatFrame", ...)
	for i = 1,NUM_CHAT_WINDOWS do
		local sFrame = "ChatFrame"..i
		local frame = _G[sFrame]
		local title,_,_,_,_,_,bShown = GetChatWindowInfo(i)
		local cmpTitle = self:LowerUTF8(title)
		if select("#",...) == 0 then
			if (not bShown and not frame.isDocked) or i == NUM_CHAT_WINDOWS then
				return sFrame,title
			end
		elseif bShown or frame.isDocked then
			for i1 = 1, select("#",...) do
				local askTitle = select(i1,...)
				if askTitle and self:LowerUTF8(askTitle) == cmpTitle then
					return sFrame,title
				end
			end
		end
	end
end

function TellTrack:EnsureChatFrame(create)		--self:Debug("EnsureChatFrame", create)
	local _chat = self._chat
	local frame = _chat.frame and _G[_chat.frame]
	if frame and frame:IsShown() then return frame end
	if not create and self._options.AutoCreate==0 then return end

	local defName,chatTitle,chatName,title = self.L["Tell Track"],_chat.title,_chat.name
	_chat.frame,title = self:FindChatFrame(defName, chatTitle, chatName);			-- get if already exists

	if not _chat.frame then
		if FCF_GetNumActiveChatFrames() >= NUM_CHAT_WINDOWS then return end			-- dont get if have to take someone elses
		title = chatName and select(2,self:DispName(chatName)) or defName
		local selected = SELECTED_DOCK_FRAME
		FCF_OpenNewWindow(title)
		FCF_SelectDockFrame(selected)
		_chat.frame = self:FindChatFrame(title)
	end
	if _chat.frame then																-- make sure clear out
		_chat.title = title
		_chat.name = _chat.name or self:NormName(title)
		frame = _G[_chat.frame]
		ChatFrame_RemoveAllMessageGroups(frame)
		ChatFrame_AddMessageGroup(frame, "WHISPER")
		ChatFrame_AddMessageGroup(frame, "BN_WHISPER")
		--ChatFrame_AddMessageGroup(frame, "BN_CONVERSATION")
		return frame
	end
end

function TellTrack:HideWhispersInOtherFrames()
	local o,sframe = self._options,self._chat.frame
	local frame = sframe and _G[sframe]
	if not frame then return end
	local func = ChatFrame_RemoveMessageGroup -- (o.HideWhispers == 1) and ChatFrame_RemoveMessageGroup or ChatFrame_AddMessageGroup
	for i = 1,NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local chatFrame = _G[sChatFrame]
		local sName,_,_,_,_,_,bShown = GetChatWindowInfo(i)
		if chatFrame and (bShown or chatFrame.isDocked) and sChatFrame ~= sframe then
			func(chatFrame, "WHISPER")
			func(chatFrame, "BN_WHISPER")
			--func(chatFrame, "BN_CONVERSATION")
		end
	end
end


function TellTrack_BNSendWhisper(presenceID, text)
	TellTrack:BNSendWhisper(presenceID, text)
end

-- why not using _INFORM event? would it show out of order?
function TellTrack:BNSendWhisper(presenceID, text)			--self:Debug("BNSendWhisper", presenceID, text)
	local target = self:GetTagForPresence(presenceID)
	if target then
		local o = self._options
		o.LastTell = self:NormName(target)
		--o.LastLang = language
		if o.Enable == 1 and strsub(text, 1, 1) ~= "<" then	-- prevent data message transfers from being used
			self:HandleMessageSentOrRecieved(target, true)
		end
	end
	if SavedBNSendWhisper then
		SavedBNSendWhisper(presenceID, text)
	end
end

function TellTrack_SendChatMessage(text, type, language, target)
	TellTrack:SendChatMessage(text, type, language, target)
end

-- why not using _INFORM event? would it show out of order?
function TellTrack:SendChatMessage(text, type, language, target)	--self:Debug("SendChatMessage", text, type, language, target)
	local o = self._options
	if type == "WHISPER" or type == "BN_WHIPSER" then
		o.LastTell = self:NormName(target)
		o.LastLang = language
		if o.Enable == 1 and strsub(text, 1, 1) ~= "<" then	-- prevent data message transfers from being used
			self:HandleMessageSentOrRecieved(target, true)
		end
	end
	if SavedSendChatMessage then
		SavedSendChatMessage(text, type, language, select(2,self:DispName(target)))
	end
end

function TellTrack_ChatFrame_OpenChat(text, chatFrame)
	local tt = TellTrack
	local _chat, _current = tt._chat, tt._current
	if text == "" and _current and _current.name and _current.name ~= ""
			and SELECTED_CHAT_FRAME:GetName() == _chat.frame then
		tt:InitiateWhisperToTarget(_current.name)
	else
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat(text, chatFrame)
		end
	end
end

function TellTrack_FCF_OpenNewWindow( name )
	if not SavedFCF_OpenNewWindow then return end
	SavedFCF_OpenNewWindow( name )
	local frame = (TellTrack._options.HideWhispers == 1) and TellTrack:FindChatFrame() or nil
	if frame and _G[frame] then
		frame = _G[frame]
		ChatFrame_RemoveMessageGroup(frame, "WHISPER")
		ChatFrame_RemoveMessageGroup(frame, "BN_WHISPER")
		--ChatFrame_RemoveMessageGroup(frame, "BN_CONVERSATION")
	end
end

function TellTrack_FCF_Close( chatFrame, fallBack )
	local _chat = TellTrack._chat
	local frame = fallBack or chatFrame or FCF_GetCurrentChatFrame()
	if frame and _chat and frame:GetName() == _chat.frame then
		_chat.frame = nil
	end
	if SavedFCF_Close then
		SavedFCF_Close(chatFrame, fallBack)
	end
end

-- universal initiate chat selection. emulates OnClick for tab
--  no way provided to select undocked tab like OnClick does, except as hack to call
--  FCFDock_SelectWindow which will fade out the main chat frame
--		if frame.isDocked then
--			FCF_SelectDockFrame(frame)
--		end
function TellTrack:SelectChatFrame(chatFrame)
	if chatFrame.isDocked then
		FCF_SelectDockFrame(chatFrame);
	else
		SELECTED_CHAT_FRAME = chatFrame;	-- required
		if GetCVar("chatStyle") ~= "classic" then
			ChatEdit_SetLastActiveWindow(chatFrame.editBox);
		end
	end
end


function TellTrack_FCFDock_SelectWindow(dock, frame)
	local tt = TellTrack
	local o,list = tt._options,tt._list

	if SavedFCFDock_SelectWindow then
		SavedFCFDock_SelectWindow(dock, frame)
	end
--print("FCFDock_SelectWindow", dock, frame and frame:GetName() or "",
--		SELECTED_CHAT_FRAME and SELECTED_CHAT_FRAME:GetName() or "")
end

function TellTrack_FCF_SelectDockFrame(frame)
	local tt = TellTrack
	local o,list = tt._options,tt._list

	if SavedFCF_SelectDockFrame then
		SavedFCF_SelectDockFrame(frame)
	end
--	print("FCF_SelectDockFrame", dock, frame and frame:GetName() or "",
--		SELECTED_CHAT_FRAME and SELECTED_CHAT_FRAME:GetName() or "")

	if frame and frame:GetName() == tt._chat.frame then
		if tt._current.showAll then
			for i = 1, o.ListSize do
				if list[i] then	list[i].unread = 0 end
			end
		elseif tt._current.name ~= "" then
			local cmpCurName = tt:CmpName(tt._current.name)
			for i = 1, o.ListSize do
				if list[i] and tt:CmpName(list[i].name) == cmpCurName then
					list[i].unread = 0
				end
			end
		end
		tt:UpdateTellTrackButtonsText()
	end
	SELECTED_CHAT_FRAME = frame;		--FIXME: is necessary
end

function TellTrack:InitiateWhisperToTarget(name)
	self:InitWhisper(select(2,self:DispName(name)), nil, self:NameType(name))
end

function TellTrack:UpdateTellTrackButtonsText()
	local o,_list = self._options,self._list
	if not _list then return end

	local listcount = #_list
	local lastToFmtStr = "|cff33ff33%s|r"
	local lastFromFmtStr = "|cffff3333%s|r"
	local noNameFmtStr = "|cff666666%s|r"

	for i = 1,TellTrackFrame.ButtonCount do
		local sline = "TellTrack" .. i
		local line = _G[sline]

		if not line then
		else
			local id = (o.ListInvert ~= 1) and ((i + o.ListOffset) - 1) or ((o.ListSize - (i + o.ListOffset)) + 2)
			local v = _list[id]

			local fmt,val = noNameFmtStr, "Empty"
			local lineText = _G[sline.."Text"],name

			if not v or not v.name then
				line.playerid = nil
				line.item = nil
				line.title = nil
			else
				line.playerid = id
				line.item = v
				line.title,val = self:DispName(v.name)
				if not TellTrack_OfflinePlayers[v.cmpName] then
					fmt = (v.sentTo) and lastToFmtStr or lastFromFmtStr
				end
				if v.unread and v.unread > 0 then
					val = "("..v.unread..") "..val
				end
				if o.LineNums == 1 then
					val = id .. ". "..val
				end
			end
			lineText:SetText(format(fmt, val))
			lineText:Show()
		end
	end
end

function TellTrack:HandleMessageSentOrRecieved( target, isSent )
	if not target then return end
	local o = self._options
	local max,_current = o.ListSize,self._current

	local cmpName,normName,dispName = (self:CmpName(target)),(self:NormName(target)),(self:DispName(target))

	-- find
	local item, id = nil, 0
	for k, v in pairs(self._list) do
		if v and v.cmpName == cmpName then item,v = v,nil; end
		if v then id = id + 1; if id ~= k then self._list[id] = v; end end		-- compress as we go
	end
	while (id >= max) do tremove(self._list, 1); id = id - 1; end
	if not item then item = { target = target, name = normName, cmpName = cmpName, dispName = dispName, unread = 0, } end
	id = id + 1; self._list[id] = item

	-- update
	item.sentTo = isSent
	if not item.unread then	item.unread = 0	end
	local chatFrame = self._chat.frame
	if not isSent and chatFrame then
--		self:Debug("sentTo",_G[chatFrame]:IsVisible(),self:CmpName(_current.name),cmpName)
		if not _G[chatFrame]:IsVisible() or (self:CmpName(_current.name) ~= cmpName and not _current.showAll) then
			item.unread = item.unread + 1
		end
	end

	-- change offset
	if o.ListInvert == 1 then
		if id < max - TellTrackFrame.ButtonCount and (max - id + 1) ~= o.ListOffset then
			self:ChangeArrayOffset(max-id+1)
		end
	else
		if id > TellTrackFrame.ButtonCount and (id-TellTrackFrame.ButtonCount+1) ~= o.ListOffset then
			self:ChangeArrayOffset(id-TellTrackFrame.ButtonCount+1)
		end
	end

	self:UpdateTellTrackButtonsText()
end

-------------------------------------------------------------

function TellTrack:GetTagForPresence(id)
	if id and type(id)=="string" then id=tonumber(id:match("(%d+)")) end
	if not id then return end
	for i = 1,BNGetNumFriends() do
		local presenceId,_,battleTag = BNGetFriendInfo(i)
		if id == presenceId then return battleTag end
	end
end

function TellTrack:GetDispForPresence(id)
	local tag = type(id)=="string" and strlower(id)
	id = tag and tonumber(id:match("(%d+)")) or nil
	for i = 1,BNGetNumFriends() do
		local presenceId,presenceName,battleTag = BNGetFriendInfo(i)
		if (id and id == presenceId) or (tag and battleTag and tag == strlower(battleTag)) then
			return presenceName end
	end
end

function TellTrack:CapUTF8( s )
	if not s then return end
	local low,up,char,t,c,d = strlower,strupper,strchar,"",32
	for i = 1,s:len() do
		d = c; c = s:byte(i);
		t = t .. (c>127 and char(c) or (((d==45 or d==32) and up or low)(char(c))))
	end
	return t
end

function TellTrack:LowerUTF8( s )
	return s and strlower(s) or s
end

function TellTrack:NormName( s )
	if s and s:find("|[Kk]") then s=self:GetTagForPresence(s) or s; return s,s end
	if s and s:find("#") then return s,s end
	if not s then return s,s end
	local o,sL = self._options
	s = self:CapUTF8(s); sL=s:gsub(o.RealmPostfix.."$","")
	if not s:find("-") then	s = sL .. o.RealmPostfix end
	return s,sL
end

function TellTrack:CmpName( s )
	if s and s:find("|[Kk]") then s=self:GetTagForPresence(s) end
	if not s then return s,s end
	local o,sL = self._options
	s = self:LowerUTF8(s);
	sL = s:gsub(o.RealmPostfixCmp.."$","")
	if not s:find("#") and not s:find("-") then s = sL .. o.RealmPostfixCmp end
	return s,sL
end

function TellTrack:DispName( s )
	if type(s) ~= "string" or s:find("|[Kk]") then return s,s end
	if s:find("#") then s = (self:GetDispForPresence(s)) or UNKNOWN; return s,s end
	local o,sL = self._options
	s = self:CapUTF8(s); sL=s:gsub(o.RealmPostfix.."$","")
	if not s:find("-") then	s = sL .. o.RealmPostfix end
	return s,sL
end

function TellTrack:NameType( s )
	return (not s or (not s:find("|[Kk]") and not s:find("#"))) and "WHISPER" or "BN_WHISPER"
end

function TellTrack:GetTimeStamp( s )
	local separator = "" -- ">"
	if ChatTimeStamps_TimeStamp and not ChatTimeStamps_SEPARATOR then
		separator = ""
	end
	if not s then
		if ChatTimeStamps_LOCALTIME then
			s = date()
		elseif Clock_GetTimeText then
			s = Clock_GetTimeText()
		else
			s = self:GetTime(nil, self._update.timediff)
		end
	end
	if type(s) == "number" then
		s = date("%H:%M:%S", s)
	end
	if type(s) ~= "string" then
		s = ""
	end
	return s..separator.." "
end
